@@ -1,7 +1,12 @@
-#include <config.h>
+#ifdef __cplusplus
+#define NULL __null
+#else
+#define NULL ((void *)0)
+#endif
 
 #include <assert.h>
 #include <ctype.h>
+#include <sys/types.h>
 #include <fcntl.h>
 #include <pthread.h>
 #include <regex.h>
