@@ -111,7 +111,8 @@ public:
  *    element or ignore the tag.
  *
  */
-KHTMLParser::KHTMLParser( KHTMLView *_parent, DocumentPtr *doc)
+KHTMLParser::KHTMLParser( KHTMLView *_parent, DocumentPtr *doc) 
+    : current(0)
 {
     //kdDebug( 6035 ) << "parser constructor" << endl;
 #if SPEED_DEBUG > 0
@@ -131,6 +132,7 @@ KHTMLParser::KHTMLParser( KHTMLView *_parent, DocumentPtr *doc)
 }
 
 KHTMLParser::KHTMLParser( DOM::DocumentFragmentImpl *i, DocumentPtr *doc )
+    : current(0)
 {
     HTMLWidget = 0;
     document = doc;
@@ -141,7 +143,7 @@ KHTMLParser::KHTMLParser( DOM::DocumentFragmentImpl *i, DocumentPtr *doc )
     blockStack = 0;
 
     reset();
-    current = i;
+    setCurrent(i);
     inBody = true;
 }
 
@@ -161,7 +163,7 @@ KHTMLParser::~KHTMLParser()
 
 void KHTMLParser::reset()
 {
-    current = document->document();
+    setCurrent(document->document());
 
     freeBlock();
 
@@ -172,7 +174,7 @@ void KHTMLParser::reset()
     haveFrameSet = false;
     haveContent = false;
     inSelect = false;
-    inStrayTableContent = false;
+    inStrayTableContent = 0;
     
     form = 0;
     map = 0;
@@ -183,6 +185,15 @@ void KHTMLParser::reset()
     discard_until = 0;
 }
 
+void KHTMLParser::setCurrent(DOM::NodeImpl *newCurrent) 
+{
+    if (newCurrent) 
+	newCurrent->ref(); 
+    if (current) 
+	current->deref(); 
+    current = newCurrent; 
+}
+
 void KHTMLParser::parseToken(Token *t)
 {
     if (t->id > 2*ID_CLOSE_TAG)
@@ -310,7 +321,7 @@ bool KHTMLParser::insertNode(NodeImpl *n, bool flat)
             if (newNode == current)
                 popBlock(id);
             else
-                current = newNode;
+                setCurrent(newNode);
 #if SPEED_DEBUG < 2
             if(!n->attached() && HTMLWidget)
                 n->attach();
@@ -345,11 +356,10 @@ bool KHTMLParser::insertNode(NodeImpl *n, bool flat)
         case ID_TR:
         case ID_TH:
         case ID_TD:
-            if (inStrayTableContent) {
+            if (inStrayTableContent && !isTableRelatedTag(current->id())) {
                 // pop out to the nearest enclosing table-related tag.
                 while (!isTableRelatedTag(current->id()))
                     popOneBlock();
-                inStrayTableContent = false;
                 return insertNode(n);
             }
             break;
@@ -405,7 +415,7 @@ bool KHTMLParser::insertNode(NodeImpl *n, bool flat)
                 DOM::NodeImpl *newNode = head->addChild(n);
                 if ( newNode ) {
                     pushBlock(id, tagPriority[id]);
-                    current = newNode;
+                    setCurrent(newNode);
 #if SPEED_DEBUG < 2
 		    if(!n->attached() && HTMLWidget)
                         n->attach();
@@ -496,6 +506,33 @@ bool KHTMLParser::insertNode(NodeImpl *n, bool flat)
                 return false;
             return true;
         }
+        case ID_CAPTION: {
+            switch (current->id()) {
+                case ID_THEAD:
+                case ID_TBODY:
+                case ID_TFOOT:
+                case ID_TR:
+                case ID_TH:
+                case ID_TD: {
+                    NodeImpl* tsection = current;
+                    if (current->id() == ID_TR)
+                        tsection = current->parent();
+                    else if (current->id() == ID_TD || current->id() == ID_TH)
+                        tsection = current->parent()->parent();
+                    NodeImpl* table = tsection->parent();
+                    int exceptioncode = 0;
+                    table->insertBefore(n, tsection, exceptioncode);
+                    pushBlock(id, tagPriority[id]);
+                    setCurrent(n);
+                    inStrayTableContent++;
+                    blockStack->strayTableContent = true;
+                    return true;
+                }
+                default:
+                    break;
+            }
+            break;
+        }
         default:
             break;
         }
@@ -577,7 +614,7 @@ bool KHTMLParser::insertNode(NodeImpl *n, bool flat)
                 popBlock(ID_TABLE); // end the table
                 handled = true;      // ...and start a new one
                 break;
-            case ID_TEXT:
+	    case ID_TEXT:
             {
                 TextImpl *t = static_cast<TextImpl *>(n);
                 if (t->containsOnlyWhitespace())
@@ -619,8 +656,8 @@ bool KHTMLParser::insertNode(NodeImpl *n, bool flat)
                         !flat && endTag[id] != DOM::FORBIDDEN)
                     {
                         pushBlock(id, tagPriority[id]);
-                        current = n;
-                        inStrayTableContent = true;
+                        setCurrent(n);
+                        inStrayTableContent++;
                         blockStack->strayTableContent = true;
                     }
                     return true;
@@ -804,10 +841,12 @@ NodeImpl *KHTMLParser::getElement(Token* t)
 
 // form elements
     case ID_FORM:
-        // close all open forms...
-        popBlock(ID_FORM);
-        form = new HTMLFormElementImpl(document);
-        n = form;
+        if (!form) {
+            // Only create a new form if we're not already inside one.
+            // This is consistent with other browsers' behavior.
+            form = new HTMLFormElementImpl(document);
+            n = form;
+        }
         break;
     case ID_BUTTON:
         n = new HTMLButtonElementImpl(document, form);
@@ -1042,6 +1081,8 @@ NodeImpl *KHTMLParser::getElement(Token* t)
     case ID_SPAN:
     case ID_NOBR:
     case ID_WBR:
+        if (t->id == ID_NOBR || t->id == ID_WBR)
+            popBlock(t->id); // Don't allow nested <nobr> or <wbr>
         n = new HTMLGenericElementImpl(document, t->id);
         break;
 
@@ -1064,7 +1105,7 @@ NodeImpl *KHTMLParser::getElement(Token* t)
         return 0;
         break;
     case ID_MARQUEE:
-        n = new HTMLGenericElementImpl(document, t->id);
+        n = new HTMLMarqueeElementImpl(document);
         break;
 // text
     case ID_TEXT:
@@ -1351,12 +1392,9 @@ void KHTMLParser::handleResidualStyleCloseTagAcrossBlocks(HTMLStackElem* elem)
             // curr->id rather than the node that you should pop to when the element gets pulled off
             // the stack.
             popOneBlock(false);
-            curr->next = 0;
             curr->node = currNode;
-            if (!residualStyleStack)
-                residualStyleStack = curr;
-            else
-                residualStyleStack->next = curr;
+            curr->next = residualStyleStack;
+            residualStyleStack = curr;
         }
         else
             popOneBlock();
@@ -1364,36 +1402,40 @@ void KHTMLParser::handleResidualStyleCloseTagAcrossBlocks(HTMLStackElem* elem)
         curr = blockStack;
     }
 
-    reopenResidualStyleTags(residualStyleStack, false); // FIXME: Deal with stray table content some day
-                                                        // if it becomes necessary to do so.
+    reopenResidualStyleTags(residualStyleStack, 0); // FIXME: Deal with stray table content some day
+                                                    // if it becomes necessary to do so.
 }
 
-void KHTMLParser::reopenResidualStyleTags(HTMLStackElem* elem, bool inMalformedTable)
+void KHTMLParser::reopenResidualStyleTags(HTMLStackElem* elem, DOM::NodeImpl* malformedTableParent)
 {
-    // Nothing required.
-    if (!elem)
-        return;
-
-    // We have some tags that need to be reopened.
-    int exceptionCode = 0;
+    // Loop for each tag that needs to be reopened.
     while (elem) {
         // Create a shallow clone of the DOM node for this element.
         NodeImpl* newNode = elem->node->cloneNode(false); 
 
-        // Append the new node.
-        current->appendChild(newNode, exceptionCode);
+        // Append the new node. In the malformed table case, we need to insert before the table,
+        // which will be the last child.
+        int exceptionCode = 0;
+        if (malformedTableParent)
+            malformedTableParent->insertBefore(newNode, malformedTableParent->lastChild(), exceptionCode);
+        else
+            current->appendChild(newNode, exceptionCode);
+        // FIXME: Is it really OK to ignore the exceptions here?
 
         // Now push a new stack element for this node we just created.
         pushBlock(elem->id, elem->level);
 
         // Set our strayTableContent boolean if needed, so that the reopened tag also knows
         // that it is inside a malformed table.
-        blockStack->strayTableContent = !inStrayTableContent && inMalformedTable;
+        blockStack->strayTableContent = malformedTableParent != 0;
         if (blockStack->strayTableContent)
-            inStrayTableContent = true;
-        
+            inStrayTableContent++;
+
+        // Clear our malformed table parent variable.
+        malformedTableParent = 0;
+
         // Update |current| manually to point to the new node.
-        current = newNode;
+        setCurrent(newNode);
         
         // Advance to the next tag that needs to be reopened.
         HTMLStackElem* next = elem->next;
@@ -1447,24 +1489,26 @@ void KHTMLParser::popBlock( int _id )
 
     bool isAffectedByStyle = isAffectedByResidualStyle(Elem->id);
     HTMLStackElem* residualStyleStack = 0;
-
-    bool residualStyleInMalformedTable = false;
+    NodeImpl* malformedTableParent = 0;
+    
     Elem = blockStack;
     while (Elem)
     {
         if (Elem->id == _id)
         {
-            bool strayTable = inStrayTableContent;
-            NodeImpl* shiftedContentParent = current ? current->parentNode() : 0;
+            int strayTable = inStrayTableContent;
             popOneBlock();
             Elem = 0;
 
-            // This element was the root of some malformed content just inside a <table>.  If
-            // we end up needing to reopen residual style tags, the root of the reopened chain
-            // must also know that it is the root of malformed content inside a <table>.
-            if (strayTable && !inStrayTableContent && residualStyleStack) {
-                residualStyleInMalformedTable = true;
-                current = shiftedContentParent;
+            // This element was the root of some malformed content just inside an implicit or
+            // explicit <tbody> or <tr>.
+            // If we end up needing to reopen residual style tags, the root of the reopened chain
+            // must also know that it is the root of malformed content inside a <tbody>/<tr>.
+            if (strayTable && (inStrayTableContent < strayTable) && residualStyleStack) {
+                NodeImpl* curr = current;
+                while (curr && curr->id() != ID_TABLE)
+                    curr = curr->parentNode();
+                malformedTableParent = curr ? curr->parentNode() : 0;
             }
         }
         else
@@ -1488,12 +1532,9 @@ void KHTMLParser::popBlock( int _id )
                 // Elem->id rather than the node that you should pop to when the element gets pulled off
                 // the stack.
                 popOneBlock(false);
-                Elem->next = 0;
+                Elem->next = residualStyleStack;
                 Elem->node = currNode;
-                if (!residualStyleStack)
-                    residualStyleStack = Elem;
-                else
-                    residualStyleStack->next = Elem;
+                residualStyleStack = Elem;
             }
             else
                 popOneBlock();
@@ -1501,7 +1542,7 @@ void KHTMLParser::popBlock( int _id )
         }
     }
 
-    reopenResidualStyleTags(residualStyleStack, residualStyleInMalformedTable);
+    reopenResidualStyleTags(residualStyleStack, malformedTableParent);
 }
 
 void KHTMLParser::popOneBlock(bool delBlock)
@@ -1530,10 +1571,10 @@ void KHTMLParser::popOneBlock(bool delBlock)
     removeForbidden(Elem->id, forbiddenTag);
 
     blockStack = Elem->next;
-    current = Elem->node;
+    setCurrent(Elem->node);
 
     if (Elem->strayTableContent)
-        inStrayTableContent = false;
+        inStrayTableContent--;
 
     if (delBlock)
         delete Elem;
@@ -1549,7 +1590,6 @@ void KHTMLParser::freeBlock()
 {
     while (blockStack)
         popOneBlock();
-    blockStack = 0;
 }
 
 void KHTMLParser::createHead()
@@ -1616,5 +1656,5 @@ void KHTMLParser::finished()
 {
     // This ensures that "current" is not left pointing to a node when the document is destroyed.
     freeBlock();
-    current = 0;
+    setCurrent(0);
 }
