@@ -26,8 +26,19 @@
 #import "KWQKURL.h"
 
 #import "KWQAssertions.h"
+#import "KWQRegExp.h"
 #import "KWQTextCodec.h"
 
+// You may have to turn this to 0 to compile without the headers for ICU installed.
+#define HAVE_ICU_LIBRARY 1
+
+#if HAVE_ICU_LIBRARY
+#import <unicode/uidna.h>
+#endif
+
+// The simple Cocoa calls to NSString, NSURL and NSData can't throw so
+// no need to block NSExceptions here.
+
 typedef enum {
     // alpha 
     SchemeFirstChar = 1 << 0,
@@ -57,6 +68,7 @@ typedef enum {
 
     // not allowed in path
     BadChar = 1 << 7
+
 } URLCharacterClasses;
 
 static const char hexDigits[17] = "0123456789ABCDEF";
@@ -190,27 +202,23 @@ static const unsigned char characterClassTable[256] = {
 
 static int copyPathRemovingDots(char *dst, const char *src, int srcStart, int srcEnd);
 
-// FIXME: convert to inline functions
-
-#define IS_SCHEME_FIRST_CHAR(c) (characterClassTable[(unsigned char)c] & SchemeFirstChar)
-#define IS_SCHEME_CHAR(c) (characterClassTable[(unsigned char)c] & SchemeChar)
-#define IS_USERINFO_CHAR(c) (characterClassTable[(unsigned char)c] & UserInfoChar)
-#define IS_HOSTNAME_CHAR(c) (characterClassTable[(unsigned char)c] & HostnameChar)
-#define IS_IPV6_CHAR(c) (characterClassTable[(unsigned char)c] & IPv6Char)
-#define IS_PATH_SEGMENT_END_CHAR(c) (characterClassTable[(unsigned char)c] & PathSegmentEndChar)
-#define IS_BAD_CHAR(c) (characterClassTable[(unsigned char)c] & BadChar)
-#define IS_HEX_DIGIT(c) (characterClassTable[(unsigned char)c] & HexDigitChar)
+static inline bool isSchemeFirstChar(unsigned char c) { return characterClassTable[c] & SchemeFirstChar; }
+static inline bool isSchemeChar(unsigned char c) { return characterClassTable[c] & SchemeChar; }
+static inline bool isUserInfoChar(unsigned char c) { return characterClassTable[c] & UserInfoChar; }
+static inline bool isHostnameChar(unsigned char c) { return characterClassTable[c] & HostnameChar; }
+static inline bool isIPv6Char(unsigned char c) { return characterClassTable[c] & IPv6Char; }
+static inline bool isPathSegmentEndChar(unsigned char c) { return characterClassTable[c] & PathSegmentEndChar; }
+static inline bool isBadChar(unsigned char c) { return characterClassTable[c] & BadChar; }
+static inline bool isHexDigit(unsigned char c) { return characterClassTable[c] & HexDigitChar; }
 
 
 // KURL
 
-KURL::KURL() :
-    m_isValid(false)
+KURL::KURL() : m_isValid(false)
 {
 }
 
-KURL::KURL(const char *url) :
-    m_isValid(true)
+KURL::KURL(const char *url)
 {
     if (url != NULL && url[0] == '/') {
 	QString qurl = QString("file:") + url;
@@ -220,9 +228,7 @@ KURL::KURL(const char *url) :
     }
 }
 
-KURL::KURL(const QString &url) :
-    m_isValid(true)
-
+KURL::KURL(const QString &url)
 {
     if (!url.isEmpty() && url[0] == '/') {
 	QString fileUrl = QString("file:") + url;
@@ -232,6 +238,26 @@ KURL::KURL(const QString &url) :
     }
 }
 
+KURL::KURL(NSURL *url)
+{
+    if (url) {
+        CFIndex bufferLength = CFURLGetBytes((CFURLRef)url, NULL, 0);
+        char *bytes = new char [bufferLength + 1];
+        CFURLGetBytes((CFURLRef)url, (UInt8 *)bytes, bufferLength);
+	bytes[bufferLength] = '\0';
+        if (bytes[0] == '/') {
+            QString fileUrl = QString("file:") + bytes;
+            parse(fileUrl.ascii(), &fileUrl);
+        } else {
+            parse(bytes, NULL);
+        }
+        delete [] bytes;
+    }
+    else {
+        parse("", NULL);
+    }
+}
+
 KURL::KURL(const KURL &base, const QString &relative, const QTextCodec *codec)
 {
     // Allow at lest absolute URLs to resolve against an empty URL.
@@ -249,8 +275,40 @@ KURL::KURL(const KURL &base, const QString &relative, const QTextCodec *codec)
         strBuffer = 0;
         str = relative.ascii();
     } else {
-        QCString decoded = codec ? codec->fromUnicode(relative)
-            : QTextCodec(kCFStringEncodingUTF8).fromUnicode(relative);
+#if HAVE_ICU_LIBRARY
+        QString s = encodeHostnames(relative);
+#else
+        QString s = relative;
+#endif
+
+        // Always use UTF-8 if the protocol is file, mailto, or help because that's
+        // what these protocols expect.
+        if (codec) {
+            QString protocol;
+            if (relative.length() > 0 && isSchemeFirstChar(relative.at(0).latin1())) {
+                for (uint i = 1; i < relative.length(); i++) {
+                    char p = relative.at(i).latin1();
+                    if (p == ':') {
+                        protocol = relative.left(i);
+                        break;
+                    }
+                    if (!isSchemeChar(p)) {
+                        break;
+                    }
+                }
+            }
+            if (!protocol) {
+                protocol = base.protocol();
+            }
+            protocol = protocol.lower();
+            if (protocol == "file" || protocol == "mailto" || protocol == "help") {
+                codec = NULL;
+            }
+        }
+
+        QCString decoded = codec
+            ? codec->fromUnicode(s)
+            : QTextCodec(kCFStringEncodingUTF8).fromUnicode(s);
         strBuffer = strdup(decoded);
         str = strBuffer;
     }
@@ -259,18 +317,18 @@ KURL::KURL(const KURL &base, const QString &relative, const QTextCodec *codec)
     // absolute URI if possible, using the "leftmost, longest"
     // algorithm. If the URI reference is absolute it will have a
     // scheme, meaning that it will have a colon before the first
-    // non-scheme element. "/", "?" and "#" are used to detect the
-    // start of a path segment, a query or a fragment ID, which would
-    // indicate no scheme had been found. IS_PATH_SEGMENT_END_CHAR
-    // tests for those three characters or NULL.
-
-    for (const char *p = str; !IS_PATH_SEGMENT_END_CHAR(*p); ++p) {
-	if (*p == ':') {
-	    absolute = true;
-	    break;
-	}
+    // non-scheme element.
+    const char *p = str;
+    if (isSchemeFirstChar(*p)) {
+        ++p;
+        while (isSchemeChar(*p)) {
+            ++p;
+        }
+        if (*p == ':') {
+            absolute = true;
+        }
     }
-    
+        
     if (absolute) {
 	parse(str, allASCII ? &relative : 0);
     } else {
@@ -284,6 +342,10 @@ KURL::KURL(const KURL &base, const QString &relative, const QTextCodec *codec)
 	if (!base.m_isValid) {
 	    QString newURL = base.urlString + str;
 	    parse(newURL.ascii(), &newURL);
+            if (strBuffer) {
+                free(strBuffer);
+            }
+            return;
 	}
 
 	switch(str[0]) {
@@ -363,14 +425,14 @@ KURL::KURL(const KURL &base, const QString &relative, const QTextCodec *codec)
 		
 		while (*relStringPos != '\0' && *relStringPos != '?' && *relStringPos != '#') {
 		    if (relStringPos[0] == '.' && bufferPos[-1] == '/') {
-			if (IS_PATH_SEGMENT_END_CHAR(relStringPos[1])) {
+			if (isPathSegmentEndChar(relStringPos[1])) {
 			    // skip over "." segment
 			    relStringPos += 1;
 			    if (relStringPos[0] == '/') {
 				relStringPos++;
 			    }
 			    continue;
-			} else if (relStringPos[1] == '.' && IS_PATH_SEGMENT_END_CHAR(relStringPos[2])) {
+			} else if (relStringPos[1] == '.' && isPathSegmentEndChar(relStringPos[2])) {
 			    // skip over ".." segment and rewind the last segment
 			    // the RFC leaves it up to the app to decide what to do with excess
 			    // ".." segments - we choose to drop them since some web content
@@ -593,6 +655,64 @@ void KURL::setPort(unsigned short i)
     }
 }
 
+void KURL::setUser(const QString &user)
+{
+    if (m_isValid) {
+        QString u;
+        int end = userEndPos;
+        if (!user.isEmpty()) {
+            // Untested code, but this is never used.
+            ASSERT_NOT_REACHED();
+#if 0
+            u = user;
+            if (userStartPos == schemeEndPos + 1) {
+                u = "//" + u;
+            }
+            // Add '@' if we didn't have one before.
+            if (end == hostEndPos || (end == passwordEndPos && urlString[end] != '@')) {
+                u += '@';
+            }
+#endif
+        } else {
+            // Remove '@' if we now have neither user nor password.
+            if (userEndPos == passwordEndPos && end != hostEndPos && urlString[end] == '@') {
+                end += 1;
+            }
+        }
+        const QString newURL = urlString.left(userStartPos) + u + urlString.mid(end);
+        parse(newURL.ascii(), &newURL);
+    }
+}
+
+void KURL::setPass(const QString &password)
+{
+    if (m_isValid) {
+        QString p;
+        int end = passwordEndPos;
+        if (!password.isEmpty()) {
+            // Untested code, but this is never used.
+            ASSERT_NOT_REACHED();
+#if 0
+            p = ':' + password + '@';
+            if (userEndPos == schemeEndPos + 1) {
+                p = "//" + p;
+            }
+            // Eat the existing '@' since we are going to add our own.
+            if (end != hostEndPos && urlString[end] == '@') {
+                end += 1;
+            }
+#endif
+        } else {
+            // Remove '@' if we now have neither user nor password.
+            if (userStartPos == userEndPos && end != hostEndPos && urlString[end] == '@') {
+                end += 1;
+            }
+        }
+        const QString newURL = urlString.left(userEndPos) + p + urlString.mid(end);
+        parse(newURL.ascii(), &newURL);
+    }
+}
+
 void KURL::setRef(const QString &s)
 {
     if (m_isValid) {
@@ -603,8 +723,8 @@ void KURL::setRef(const QString &s)
 
 void KURL::setQuery(const QString &query, int encoding_hint)
 {
-    QString q;
     if (m_isValid) {
+        QString q;
 	if (!query.isEmpty() && query[0] != '?') {
 	    q = "?" + query;
 	} else {
@@ -732,8 +852,8 @@ static void appendEscapingBadChars(char*& buffer, const char *strStart, size_t l
     const char *strEnd = strStart + length;
     while (str < strEnd) {
 	unsigned char c = *str++;
-        if (IS_BAD_CHAR(c)) {
-            if (c == '%' && strEnd - str >= 2 && IS_HEX_DIGIT(str[0]) && IS_HEX_DIGIT(str[1])) {
+        if (isBadChar(c)) {
+            if (c == '%' && strEnd - str >= 2 && isHexDigit(str[0]) && isHexDigit(str[1])) {
                 *p++ = c;
                 *p++ = *str++;
                 *p++ = *str++;
@@ -759,7 +879,17 @@ static int copyPathRemovingDots(char *dst, const char *src, int srcStart, int sr
     const char *baseStringEnd = src + srcEnd;
     char *bufferPathStart = dst;
     const char *baseStringPos = baseStringStart;
-    
+
+    // this code is unprepared for paths that do not begin with a
+    // slash and we should always have one in the source string (but a
+    // totally empty path is OK and does not need to start with a slash)
+    ASSERT(srcStart == srcEnd || baseStringPos[0] == '/');
+
+    // copy the leading slash into the destination
+    *dst = *baseStringPos;
+    baseStringPos++;
+    dst++;
+
     while (baseStringPos < baseStringEnd) {
         if (baseStringPos[0] == '.' && dst[-1] == '/') {
             if (baseStringPos[1] == '/' || baseStringPos + 1 == baseStringEnd) {
@@ -808,7 +938,7 @@ void KURL::parse(const char *url, const QString *originalString)
 	return;
     }
 
-    if (!IS_SCHEME_FIRST_CHAR(url[0])) {
+    if (!isSchemeFirstChar(url[0])) {
 	// scheme must start with an alphabetic character
 	m_isValid = false;
 	urlString = url;
@@ -817,7 +947,7 @@ void KURL::parse(const char *url, const QString *originalString)
 
     int schemeEnd = 0;
  
-    while (IS_SCHEME_CHAR(url[schemeEnd])) {
+    while (isSchemeChar(url[schemeEnd])) {
 	schemeEnd++;
     }
 
@@ -846,7 +976,7 @@ void KURL::parse(const char *url, const QString *originalString)
 	userEnd = userStart;
 
 	int colonPos = 0;
-	while (IS_USERINFO_CHAR(url[userEnd])) {
+	while (isUserInfoChar(url[userEnd])) {
 	    if (url[userEnd] == ':' && colonPos == 0) {
 		colonPos = userEnd;
 	    }
@@ -863,7 +993,7 @@ void KURL::parse(const char *url, const QString *originalString)
 		passwordStart = passwordEnd = userEnd;
 	    }
 	    hostStart = passwordEnd + 1;
-	} else if (url[userEnd] == '\0' || url[userEnd] == '[' || IS_PATH_SEGMENT_END_CHAR(url[userEnd])) {
+	} else if (url[userEnd] == '[' || isPathSegmentEndChar(url[userEnd])) {
 	    // hit the end of the authority, must have been no user
 	    // or looks like an IPv6 hostname
 	    // either way, try to parse it as a hostname
@@ -882,7 +1012,7 @@ void KURL::parse(const char *url, const QString *originalString)
 	// IPV6 IP address
 	if (url[hostEnd] == '[') {
 	    hostEnd++;
-	    while (IS_IPV6_CHAR(url[hostEnd])) {
+	    while (isIPv6Char(url[hostEnd])) {
 		hostEnd++;
 	    }
 	    if (url[hostEnd] == ']') {
@@ -894,7 +1024,7 @@ void KURL::parse(const char *url, const QString *originalString)
 		return;
 	    }
 	} else {
-	    while (IS_HOSTNAME_CHAR(url[hostEnd])) {
+	    while (isHostnameChar(url[hostEnd])) {
 		hostEnd++;
 	    }
 	}
@@ -911,7 +1041,7 @@ void KURL::parse(const char *url, const QString *originalString)
 	    portStart = portEnd = hostEnd;
 	}
 
-	if (url[portEnd] != '\0' && !IS_PATH_SEGMENT_END_CHAR(url[portEnd])) {
+	if (!isPathSegmentEndChar(url[portEnd])) {
 	    // invalid character
 	    m_isValid = false;
 	    urlString = url;
@@ -927,26 +1057,45 @@ void KURL::parse(const char *url, const QString *originalString)
 	
     int pathStart = portEnd;
     int pathEnd = pathStart;
-    while (url[pathEnd] != '\0' && url[pathEnd] != '?' && url[pathEnd] != '#') {
-        pathEnd++;
-    }
+    int queryStart;
+    int queryEnd;
+    int fragmentStart;
+    int fragmentEnd;
 
-    int queryStart = pathEnd;
-    int queryEnd = queryStart;
-    if (url[queryStart] == '?') {
-        while (url[queryEnd] != '\0' && url[queryEnd] != '#') {
+    if (!hierarchical) {
+        while (url[pathEnd] != '\0' && url[pathEnd] != '?') {
+            pathEnd++;
+        }
+    	queryStart = queryEnd = pathEnd;
+
+        while (url[queryEnd] != '\0') {
             queryEnd++;
         }
-    }
 
-    int fragmentStart = queryEnd;
-    int fragmentEnd = fragmentStart;
-    if (url[fragmentStart] == '#') {
-	fragmentStart++;
-	fragmentEnd = fragmentStart;
-	while(url[fragmentEnd] != '\0') {
-	    fragmentEnd++;
-	}
+    	fragmentStart = fragmentEnd = queryEnd;
+    }
+    else {
+        while (url[pathEnd] != '\0' && url[pathEnd] != '?' && url[pathEnd] != '#') {
+            pathEnd++;
+        }
+    
+        queryStart = pathEnd;
+        queryEnd = queryStart;
+        if (url[queryStart] == '?') {
+            while (url[queryEnd] != '\0' && url[queryEnd] != '#') {
+                queryEnd++;
+            }
+        }
+        
+        fragmentStart = queryEnd;
+        fragmentEnd = fragmentStart;
+        if (url[fragmentStart] == '#') {
+            fragmentStart++;
+            fragmentEnd = fragmentStart;
+            while(url[fragmentEnd] != '\0') {
+                fragmentEnd++;
+            }
+        }
     }
 
     // assemble it all, remembering the real ranges
@@ -1076,7 +1225,7 @@ void KURL::parse(const char *url, const QString *originalString)
        
     // add path, escaping bad characters
     
-    if (strstr(url, "/.") || strstr(url, "..")) {
+    if (hierarchical && (strstr(url, "/.") || strstr(url, ".."))) {
         char static_path_buffer[4096];
         char *path_buffer;
         uint pathBufferLength = pathEnd - pathStart + 1;
@@ -1170,7 +1319,7 @@ QString KURL::encode_string(const QString& notEncodedString)
     const char *strEnd = str + asUTF8.length();
     while (str < strEnd) {
 	unsigned char c = *str++;
-        if (IS_BAD_CHAR(c)) {
+        if (isBadChar(c)) {
             *p++ = '%';
             *p++ = hexDigits[c >> 4];
             *p++ = hexDigits[c & 0xF];
@@ -1189,3 +1338,199 @@ QString KURL::encode_string(const QString& notEncodedString)
 
     return result;
 }
+
+NSURL *KURL::getNSURL() const
+{
+    const UInt8 *bytes = (const UInt8 *)(urlString.latin1());
+    NSURL *result = nil;
+    if (urlString.length() > 0) {
+        // NOTE: We use UTF-8 here since this encoding is used when computing strings when returning URL components
+        // (e.g calls to NSURL -path). However, this function is not tolerant of illegal UTF-8 sequences, which
+        // could either be a malformed string or bytes in a different encoding, like shift-jis, so we fall back
+        // onto using ISO Latin 1 in those cases.
+        result = (NSURL *)CFURLCreateAbsoluteURLWithBytes(NULL, bytes, urlString.length(), kCFStringEncodingUTF8, NULL, TRUE);
+        if (!result) {
+            result = (NSURL *)CFURLCreateAbsoluteURLWithBytes(NULL, bytes, urlString.length(), kCFStringEncodingISOLatin1, NULL, TRUE);
+        }
+        [result autorelease];
+    }
+    else {
+        result = [NSURL URLWithString:@""];
+    }
+    
+    return result;
+}
+
+NSData *KURL::getNSData() const
+{
+    const UInt8 *bytes = (const UInt8 *)(urlString.latin1());
+    return [NSData dataWithBytes:bytes length:urlString.length()];
+}
+
+#if HAVE_ICU_LIBRARY
+
+QString KURL::encodeHostnames(const QString &s)
+{
+    if (s.startsWith("mailto:", false)) {
+        const QMemArray<KWQIntegerPair> hostnameRanges = findHostnamesInMailToURL(s);
+        uint n = hostnameRanges.size();
+        if (n != 0) {
+            QString result;
+            uint p = 0;
+            for (uint i = 0; i < n; ++i) {
+                const KWQIntegerPair &r = hostnameRanges[i];
+                result += s.mid(p, r.start);
+                result += encodeHostname(s.mid(r.start, r.end - r.start));
+                p = r.end;
+            }
+            result += s.mid(p);
+            return result;
+        }
+    } else {
+        int hostStart, hostEnd;
+        if (findHostnameInHierarchicalURL(s, hostStart, hostEnd)) {
+            return s.left(hostStart) + encodeHostname(s.mid(hostStart, hostEnd - hostStart)) + s.mid(hostEnd); 
+        }
+    }
+    return s;
+}
+
+bool KURL::findHostnameInHierarchicalURL(const QString &s, int &startOffset, int &endOffset)
+{
+    // Find the host name in a hierarchical URL.
+    // It comes after a "://" sequence, with scheme characters preceding.
+    // If ends with the end of the string or a ":" or a path segment ending character.
+    // If there is a "@" character, the host part is just the part after the "@".
+    int separator = s.find("://");
+    if (separator <= 0) {
+        return false;
+    }
+
+    // Check that all characters before the :// are valid scheme characters.
+    if (!isSchemeFirstChar(s[0].latin1())) {
+        return false;
+    }
+    for (int i = 1; i < separator; ++i) {
+        if (!isSchemeChar(s[i].latin1())) {
+            return false;
+        }
+    }
+
+    // Start after the separator.
+    int authorityStart = separator + 3;
+
+    // Find terminating character.
+    int length = s.length();
+    int hostnameEnd = length;
+    for (int i = authorityStart; i < length; ++i) {
+        char c = s[i].latin1();
+        if (c == ':' || (isPathSegmentEndChar(c) && c != '\0')) {
+            hostnameEnd = i;
+            break;
+        }
+    }
+
+    // Find "@" for the start of the host name.
+    int userInfoTerminator = s.find('@', authorityStart);
+    int hostnameStart;
+    if (userInfoTerminator == -1 || userInfoTerminator > hostnameEnd) {
+        hostnameStart = authorityStart;
+    } else {
+        hostnameStart = userInfoTerminator + 1;
+    }
+
+    startOffset = hostnameStart;
+    endOffset = hostnameEnd;
+    return true;
+}
+
+QString KURL::encodeHostname(const QString &s)
+{
+    // Needs to be big enough to hold an IDN-encoded name.
+    // For host names bigger than this, we won't do IDN encoding, which is almost certainly OK.
+    const unsigned hostnameBufferLength = 2048;
+
+    if (s.isAllASCII() || s.length() > hostnameBufferLength) {
+        return s;
+    }
+
+    UChar buffer[hostnameBufferLength];    
+    UErrorCode error = U_ZERO_ERROR;
+    int32_t numCharactersConverted = uidna_IDNToASCII
+        (reinterpret_cast<const UChar *>(s.unicode()), s.length(), buffer, hostnameBufferLength, UIDNA_ALLOW_UNASSIGNED, NULL, &error);
+    if (error != U_ZERO_ERROR) {
+        return s;
+    }
+    return QString(reinterpret_cast<QChar *>(buffer), numCharactersConverted);
+}
+
+QMemArray<KWQIntegerPair> KURL::findHostnamesInMailToURL(const QString &s)
+{
+    // In a mailto: URL, host names come after a '@' character and end with a '>' or ',' or '?' or end of string character.
+    // Skip quoted strings so that characters in them don't confuse us.
+    // When we find a '?' character, we are past the part of the URL that contains host names.
+
+    QMemArray<KWQIntegerPair> a;
+
+    int p = 0;
+    while (1) {
+        // Find start of host name or of quoted string.
+        int hostnameOrStringStart = s.find(QRegExp("[\"@?]"), p);
+        if (hostnameOrStringStart == -1) {
+            return a;
+        }
+        QChar c = s[hostnameOrStringStart];
+        p = hostnameOrStringStart + 1;
+
+        if (c == '?') {
+            return a;
+        }
+        
+        if (c == '@') {
+            // Find end of host name.
+            int hostnameStart = p;
+            int hostnameEnd = s.find(QRegExp("[>,?]"), p);
+            bool done;
+            if (hostnameEnd == -1) {
+                hostnameEnd = s.length();
+                done = true;
+            } else {
+                p = hostnameEnd;
+                done = false;
+            }
+
+            int i = a.size();
+            a.resize(i + 1);
+            a[i] = KWQIntegerPair(hostnameStart, hostnameEnd);
+
+            if (done) {
+                return a;
+            }
+        } else {
+            // Skip quoted string.
+            ASSERT(c == '"');
+            while (1) {
+                int escapedCharacterOrStringEnd = s.find(QRegExp("[\"\\]"), p);
+                if (escapedCharacterOrStringEnd == -1) {
+                    return a;
+                }
+                c = s[escapedCharacterOrStringEnd];
+                p = escapedCharacterOrStringEnd + 1;
+                
+                // If we are the end of the string, then break from the string loop back to the host name loop.
+                if (c == '"') {
+                    break;
+                }
+                
+                // Skip escaped character.
+                ASSERT(c == '\\');
+                if (p == static_cast<int>(s.length())) {
+                    return a;
+                }                
+                ++p;
+            }
+        }
+    }
+}
+
+#endif // HAVE_ICU_LIBRARY
